@@ -22,10 +22,8 @@
 #ifndef AP_RELEASE_H
 #define AP_RELEASE_H
 
-#include "apr_general.h" /* stringify */
-
 #define AP_SERVER_COPYRIGHT \
-    "Copyright 2011 The Apache Software Foundation."
+  "Copyright 2017 The Apache Software Foundation."
 
 /*
  * The below defines the base string of the Server: header. Additional
@@ -34,24 +32,36 @@
  * The tokens are listed in order of their significance for identifying the
  * application.
  *
- * "Product tokens should be short and to the point -- use of them for 
+ * "Product tokens should be short and to the point -- use of them for
  * advertizing or other non-essential information is explicitly forbidden."
  *
- * Example: "Apache/1.1.0 MrWidget/0.1-alpha" 
+ * Example: "Apache/1.1.0 MrWidget/0.1-alpha"
  */
 #define AP_SERVER_BASEVENDOR "Apache Software Foundation"
 #define AP_SERVER_BASEPROJECT "Apache HTTP Server"
 #define AP_SERVER_BASEPRODUCT "Apache"
 
 #define AP_SERVER_MAJORVERSION_NUMBER 2
-#define AP_SERVER_MINORVERSION_NUMBER 2
-#define AP_SERVER_PATCHLEVEL_NUMBER   19
+#define AP_SERVER_MINORVERSION_NUMBER 4
+#define AP_SERVER_PATCHLEVEL_NUMBER   29
 #define AP_SERVER_DEVBUILD_BOOLEAN    0
 
-#if AP_SERVER_DEVBUILD_BOOLEAN
-#define AP_SERVER_ADD_STRING          "-dev"
-#else
+/* Synchronize the above with docs/manual/style/version.ent */
+
+#if !AP_SERVER_DEVBUILD_BOOLEAN
 #define AP_SERVER_ADD_STRING          ""
+#else
+#ifndef AP_SERVER_ADD_STRING
+#define AP_SERVER_ADD_STRING          "-dev"
+#endif
+#endif
+
+/* APR_STRINGIFY is defined here, and also in apr_general.h, so wrap it */
+#ifndef APR_STRINGIFY
+/** Properly quote a value as a string in the C preprocessor */
+#define APR_STRINGIFY(n) APR_STRINGIFY_HELPER(n)
+/** Helper macro for APR_STRINGIFY */
+#define APR_STRINGIFY_HELPER(n) #n
 #endif
 
 /* keep old macros as well */
@@ -66,8 +76,8 @@
 #define AP_SERVER_VERSION       AP_SERVER_BASEVERSION
 
 /* macro for Win32 .rc files using numeric csv representation */
-#define AP_SERVER_PATCHLEVEL_CSV AP_SERVER_MAJORVERSION_NUMBER ##, \
-                               ##AP_SERVER_MINORVERSION_NUMBER ##, \
-                               ##AP_SERVER_PATCHLEVEL_NUMBER
+#define AP_SERVER_PATCHLEVEL_CSV AP_SERVER_MAJORVERSION_NUMBER, \
+                                 AP_SERVER_MINORVERSION_NUMBER, \
+                                 AP_SERVER_PATCHLEVEL_NUMBER
 
 #endif
