@@ -1,9 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -321,10 +321,14 @@ static int dav_rw_liveprop(dav_propdb *propdb, dav_elem_private *priv)
 /* do a sub-request to fetch properties for the target resource's URI. */
 static void dav_do_prop_subreq(dav_propdb *propdb)
 {
+    /* need to escape the uri that's in the resource struct because during
+     * the property walker it's not encoded. */
+    const char *e_uri = ap_escape_uri(propdb->resource->pool,
+                                      propdb->resource->uri);
+
     /* perform a "GET" on the resource's URI (note that the resource
        may not correspond to the current request!). */
-    propdb->subreq = ap_sub_req_lookup_uri(propdb->resource->uri, propdb->r,
-                                           NULL);
+    propdb->subreq = ap_sub_req_lookup_uri(e_uri, propdb->r, NULL);
 }
 
 static dav_error * dav_insert_coreprop(dav_propdb *propdb,
@@ -479,12 +483,12 @@ static void dav_output_prop_name(apr_pool_t *pool,
     apr_text_append(pool, phdr, s);
 }
 
-static void dav_insert_xmlns(apr_pool_t *p, const char *pre_prefix, int ns,
+static void dav_insert_xmlns(apr_pool_t *p, const char *pre_prefix, long ns,
                              const char *ns_uri, apr_text_header *phdr)
 {
     const char *s;
 
-    s = apr_psprintf(p, " xmlns:%s%d=\"%s\"", pre_prefix, ns, ns_uri);
+    s = apr_psprintf(p, " xmlns:%s%ld=\"%s\"", pre_prefix, ns, ns_uri);
     apr_text_append(p, phdr, s);
 }
 
@@ -526,14 +530,14 @@ DAV_DECLARE(dav_error *)dav_open_propdb(request_rec *r, dav_lockdb *lockdb,
 
 #if DAV_DEBUG
     if (resource->uri == NULL) {
-        return dav_new_error(r->pool, HTTP_INTERNAL_SERVER_ERROR, 0,
+        return dav_new_error(r->pool, HTTP_INTERNAL_SERVER_ERROR, 0, 0,
                              "INTERNAL DESIGN ERROR: resource must define "
                              "its URI.");
     }
 #endif
 
     propdb->r = r;
-    propdb->p = r->pool; /* ### get rid of this */
+    apr_pool_create(&propdb->p, r->pool);
     propdb->resource = resource;
     propdb->ns_xlate = ns_xlate;
 
@@ -554,10 +558,14 @@ DAV_DECLARE(dav_error *)dav_open_propdb(request_rec *r, dav_lockdb *lockdb,
 
 DAV_DECLARE(void) dav_close_propdb(dav_propdb *propdb)
 {
-    if (propdb->db == NULL)
-        return;
+    if (propdb->db != NULL) {
+        (*propdb->db_hooks->close)(propdb->db);
+    }
 
-    (*propdb->db_hooks->close)(propdb->db);
+    /* Currently, mod_dav's pool usage doesn't allow clearing this pool. */
+#if 0
+    apr_pool_destroy(propdb->p);
+#endif
 }
 
 DAV_DECLARE(dav_get_props_result) dav_get_allprops(dav_propdb *propdb,
@@ -589,13 +597,14 @@ DAV_DECLARE(dav_get_props_result) dav_get_allprops(dav_propdb *propdb,
         if (propdb->db != NULL) {
             dav_xmlns_info *xi = dav_xmlns_create(propdb->p);
             dav_prop_name name;
+            dav_error *err;
 
             /* define (up front) any namespaces the db might need */
             (void) (*db_hooks->define_namespaces)(propdb->db, xi);
 
             /* get the first property name, beginning the scan */
-            (void) (*db_hooks->first_name)(propdb->db, &name);
-            while (name.ns != NULL) {
+            err = (*db_hooks->first_name)(propdb->db, &name);
+            while (!err && name.ns) {
 
                 /*
                 ** We also look for <DAV:getcontenttype> and
@@ -614,7 +623,6 @@ DAV_DECLARE(dav_get_props_result) dav_get_allprops(dav_propdb *propdb,
                 }
 
                 if (what == DAV_PROP_INSERT_VALUE) {
-                    dav_error *err;
                     int found;
 
                     if ((err = (*db_hooks->output_value)(propdb->db, &name,
@@ -633,7 +641,7 @@ DAV_DECLARE(dav_get_props_result) dav_get_allprops(dav_propdb *propdb,
                 }
 
               next_key:
-                (void) (*db_hooks->next_name)(propdb->db, &name);
+                err = (*db_hooks->next_name)(propdb->db, &name);
             }
 
             /* all namespaces have been entered into xi. generate them into
@@ -760,7 +768,7 @@ DAV_DECLARE(dav_get_props_result) dav_get_props(dav_propdb *propdb,
                     for (scan_ns_uri = priv->provider->namespace_uris;
                          *scan_ns_uri != NULL;
                          ++scan_ns_uri) {
-                        int ns;
+                        long ns;
 
                         ns = dav_get_liveprop_ns_index(*scan_ns_uri);
                         if (marks_liveprop[ns])
@@ -784,6 +792,7 @@ DAV_DECLARE(dav_get_props_result) dav_get_props(dav_propdb *propdb,
 #if 0
                 /* ### need to change signature to return an error */
                 return dav_new_error(propdb->p, HTTP_INTERNAL_SERVER_ERROR, 0,
+                                     0,
                                      "INTERNAL DESIGN ERROR: insert_liveprop "
                                      "did not insert what was asked for.");
 #endif
@@ -927,7 +936,7 @@ DAV_DECLARE_NONSTD(void) dav_prop_validate(dav_prop_ctx *ctx)
 
     if (!dav_rw_liveprop(propdb, priv)) {
         ctx->err = dav_new_error(propdb->p, HTTP_CONFLICT,
-                                 DAV_ERR_PROP_READONLY,
+                                 DAV_ERR_PROP_READONLY, 0,
                                  "Property is read-only.");
         return;
     }
@@ -963,7 +972,7 @@ DAV_DECLARE_NONSTD(void) dav_prop_validate(dav_prop_ctx *ctx)
     */
     if (propdb->db == NULL) {
         ctx->err = dav_new_error(propdb->p, HTTP_INTERNAL_SERVER_ERROR,
-                                 DAV_ERR_PROP_NO_DATABASE,
+                                 DAV_ERR_PROP_NO_DATABASE, 0,
                                  "Attempted to set/remove a property "
                                  "without a valid, open, read/write "
                                  "property database.");
