@@ -25,7 +25,7 @@
 #define APR_WANT_MEMFUNC
 #define APR_WANT_STRFUNC
 #include "apr_strings.h"
-#include "apr_hooks.h"
+#include "ap_hooks.h"
 #include "apr_optional_hooks.h"
 #include "apr_buckets.h"
 
@@ -37,13 +37,13 @@
 #include "util_script.h"
 
 #include "mod_proxy.h"
+#include "scgi.h"
 
 
 #define SCHEME "scgi"
 #define PROXY_FUNCTION "SCGI"
 #define SCGI_MAGIC "SCGI"
 #define SCGI_PROTOCOL_VERSION "1"
-#define SCGI_DEFAULT_PORT (4000)
 
 /* just protect from typos */
 #define CONTENT_LENGTH "CONTENT_LENGTH"
@@ -64,10 +64,12 @@ typedef struct {
 
 const char *scgi_sendfile_off = "off";
 const char *scgi_sendfile_on = "X-Sendfile";
+const char *scgi_internal_redirect_off = "off";
+const char *scgi_internal_redirect_on = "Location";
 
 typedef struct {
     const char *sendfile;
-    int internal_redirect;
+    const char *internal_redirect;
 } scgi_config;
 
 
@@ -143,7 +145,7 @@ static const apr_bucket_type_t bucket_type_socket_ex = {
     "SOCKET_EX", 5, APR_BUCKET_DATA,
     apr_bucket_destroy_noop,
     bucket_socket_ex_read,
-    apr_bucket_setaside_notimpl, 
+    apr_bucket_setaside_notimpl,
     apr_bucket_split_notimpl,
     apr_bucket_copy_notimpl
 };
@@ -176,22 +178,29 @@ static int scgi_canon(request_rec *r, char *url)
 {
     char *host, sport[sizeof(":65535")];
     const char *err, *path;
-    apr_port_t port = SCGI_DEFAULT_PORT;
+    apr_port_t port, def_port;
 
     if (strncasecmp(url, SCHEME "://", sizeof(SCHEME) + 2)) {
         return DECLINED;
     }
     url += sizeof(SCHEME); /* Keep slashes */
 
+    port = def_port = SCGI_DEF_PORT;
+
     err = ap_proxy_canon_netloc(r->pool, &url, NULL, NULL, &host, &port);
     if (err) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00857)
                       "error parsing URL %s: %s", url, err);
         return HTTP_BAD_REQUEST;
     }
-        
-    apr_snprintf(sport, sizeof(sport), ":%u", port);
-        
+
+    if (port != def_port) {
+        apr_snprintf(sport, sizeof(sport), ":%u", port);
+    }
+    else {
+        sport[0] = '\0';
+    }
+
     if (ap_strchr(host, ':')) { /* if literal IPv6 address */
         host = apr_pstrcat(r->pool, "[", host, "]", NULL);
     }
@@ -204,7 +213,11 @@ static int scgi_canon(request_rec *r, char *url)
 
     r->filename = apr_pstrcat(r->pool, "proxy:" SCHEME "://", host, sport, "/",
                               path, NULL);
-    r->path_info = apr_pstrcat(r->pool, "/", path, NULL);
+
+    if (apr_table_get(r->subprocess_env, "proxy-scgi-pathinfo")) {
+        r->path_info = apr_pstrcat(r->pool, "/", path, NULL);
+    }
+
     return OK;
 }
 
@@ -221,9 +234,9 @@ static int sendall(proxy_conn_rec *conn, const char *buf, apr_size_t length,
     while (length > 0) {
         written = length;
         if ((rv = apr_socket_send(conn->sock, buf, &written)) != APR_SUCCESS) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
-                          "proxy: " PROXY_FUNCTION ": sending data to "
-                          "%s:%u failed", conn->hostname, conn->port);
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(00858)
+                          "sending data to %s:%u failed",
+                          conn->hostname, conn->port);
             return HTTP_SERVICE_UNAVAILABLE;
         }
 
@@ -266,7 +279,7 @@ static int send_headers(request_rec *r, proxy_conn_rec *conn)
      */
     env_table = apr_table_elts(r->subprocess_env);
     env = (apr_table_entry_t *)env_table->elts;
-    for (j=0; j<env_table->nelts; ++j) {
+    for (j = 0; j < env_table->nelts; ++j) {
         if (   (!strcmp(env[j].key, GATEWAY_INTERFACE))
             || (!strcmp(env[j].key, CONTENT_LENGTH))
             || (!strcmp(env[j].key, SCGI_MAGIC))) {
@@ -294,7 +307,7 @@ static int send_headers(request_rec *r, proxy_conn_rec *conn)
     memcpy(cp, SCGI_PROTOCOL_VERSION, sizeof(SCGI_PROTOCOL_VERSION));
     cp += sizeof(SCGI_PROTOCOL_VERSION);
 
-    for (j=0; j<env_table->nelts; ++j) {
+    for (j = 0; j < env_table->nelts; ++j) {
         if (   (!strcmp(env[j].key, GATEWAY_INTERFACE))
             || (!strcmp(env[j].key, CONTENT_LENGTH))
             || (!strcmp(env[j].key, SCGI_MAGIC))) {
@@ -332,9 +345,8 @@ static int send_request_body(request_rec *r, proxy_conn_rec *conn)
             readlen = ap_get_client_block(r, buf, AP_IOBUFSIZE);
         }
         if (readlen == -1) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-                          "proxy: " PROXY_FUNCTION ": receiving request body "
-                          "failed");
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00859)
+                          "receiving request body failed");
             return HTTP_INTERNAL_SERVER_ERROR;
         }
     }
@@ -365,11 +377,12 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
     b = apr_bucket_eos_create(r->connection->bucket_alloc);
     APR_BRIGADE_INSERT_TAIL(bb, b);
 
-    status = ap_scan_script_header_err_brigade(r, bb, NULL);
+    status = ap_scan_script_header_err_brigade_ex(r, bb, NULL,
+                                                  APLOG_MODULE_INDEX);
     if (status != OK) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-                      "proxy: " PROXY_FUNCTION ": error reading response "
-                      "headers from %s:%u", conn->hostname, conn->port);
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00860)
+                      "error reading response headers from %s:%u",
+                      conn->hostname, conn->port);
         r->status_line = NULL;
         apr_brigade_destroy(bb);
         return status;
@@ -387,9 +400,8 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
         if (location) {
             scgi_request_config *req_conf = apr_palloc(r->pool,
                                                        sizeof(*req_conf));
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
-                          "proxy: " PROXY_FUNCTION ": Found %s: %s - "
-                          "preparing subrequest.",
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00861)
+                          "Found %s: %s - preparing subrequest.",
                           conf->sendfile, location);
 
             if (err) {
@@ -407,11 +419,29 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
         }
     }
 
-    if (conf->internal_redirect && r->status == HTTP_OK) {
-        location = apr_table_get(r->headers_out, "Location");
+    if (r->status == HTTP_OK 
+        && (!conf->internal_redirect /* default === On */
+            || conf->internal_redirect != scgi_internal_redirect_off)) {
+        short err = 1;
+        const char *location_header = conf->internal_redirect ? 
+            conf->internal_redirect : scgi_internal_redirect_on;
+
+        location = apr_table_get(r->err_headers_out, location_header);
+        if (!location) {
+            err = 0;
+            location = apr_table_get(r->headers_out, location_header);
+        }
         if (location && *location == '/') {
             scgi_request_config *req_conf = apr_palloc(r->pool,
                                                        sizeof(*req_conf));
+            if (strcasecmp(location_header, "Location")) {
+                if (err) {
+                    apr_table_unset(r->err_headers_out, location_header);
+                }
+                else {
+                    apr_table_unset(r->headers_out, location_header);
+                }
+            }
             req_conf->location = location;
             req_conf->type = scgi_internal_redirect;
             ap_set_module_config(r->request_config, &proxy_scgi_module,
@@ -421,8 +451,9 @@ static int pass_response(request_rec *r, proxy_conn_rec *conn)
         }
     }
 
-    /* XXX: What could we do with that return code? */
-    (void)ap_pass_brigade(r->output_filters, bb);
+    if (ap_pass_brigade(r->output_filters, bb)) {
+        return AP_FILTER_ERROR;
+    }
 
     return OK;
 }
@@ -439,9 +470,8 @@ static int scgi_request_status(int *status, request_rec *r)
                                             &proxy_scgi_module))) {
         switch (req_conf->type) {
         case scgi_internal_redirect:
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
-                          "proxy: " PROXY_FUNCTION ": Internal redirect to %s",
-                          req_conf->location);
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00862)
+                          "Internal redirect to %s", req_conf->location);
 
             r->status_line = NULL;
             if (r->method_number != M_GET) {
@@ -455,15 +485,14 @@ static int scgi_request_status(int *status, request_rec *r)
             /* break; */
 
         case scgi_sendfile:
-            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
-                          "proxy: " PROXY_FUNCTION ": File subrequest to %s",
-                          req_conf->location);
+            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00863)
+                          "File subrequest to %s", req_conf->location);
             do {
                 request_rec *rr;
 
                 rr = ap_sub_req_lookup_file(req_conf->location, r,
                                             r->output_filters);
-                if (rr->status == HTTP_OK && rr->finfo.filetype != 0) {
+                if (rr->status == HTTP_OK && rr->finfo.filetype != APR_NOFILE) {
                     /*
                      * We don't touch Content-Length here. It might be
                      * borked (there's plenty of room for a race condition).
@@ -472,7 +501,7 @@ static int scgi_request_status(int *status, request_rec *r)
                     ap_run_sub_req(rr);
                 }
                 else {
-                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00864)
                                   "Subrequest to file '%s' not possible. "
                                   "(rr->status=%d, rr->finfo.filetype=%d)",
                                   req_conf->location, rr->status,
@@ -480,7 +509,7 @@ static int scgi_request_status(int *status, request_rec *r)
                     *status = HTTP_INTERNAL_SERVER_ERROR;
                     return *status;
                 }
-            } while(0);
+            } while (0);
 
             return OK;
             /* break; */
@@ -501,16 +530,15 @@ static int scgi_handler(request_rec *r, proxy_worker *worker,
     int status;
     proxy_conn_rec *backend = NULL;
     apr_pool_t *p = r->pool;
-    apr_uri_t *uri = apr_palloc(r->pool, sizeof(*uri));
+    apr_uri_t *uri;
     char dummy;
 
     if (strncasecmp(url, SCHEME "://", sizeof(SCHEME) + 2)) {
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
-                      "proxy: " PROXY_FUNCTION ": declining URL %s", url);
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, APLOGNO(00865)
+                      "declining URL %s", url);
         return DECLINED;
     }
-    url += sizeof(SCHEME); /* keep the slashes */
-    
+
     /* Create space for state information */
     status = ap_proxy_acquire_connection(PROXY_FUNCTION, &backend, worker,
                                          r->server);
@@ -520,6 +548,7 @@ static int scgi_handler(request_rec *r, proxy_worker *worker,
     backend->is_ssl = 0;
 
     /* Step One: Determine Who To Connect To */
+    uri = apr_palloc(p, sizeof(*uri));
     status = ap_proxy_determine_connection(p, r, conf, worker, backend,
                                            uri, &url, proxyname, proxyport,
                                            &dummy, 1);
@@ -529,9 +558,9 @@ static int scgi_handler(request_rec *r, proxy_worker *worker,
 
     /* Step Two: Make the Connection */
     if (ap_proxy_connect_backend(PROXY_FUNCTION, backend, worker, r->server)) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
-                     "proxy: " PROXY_FUNCTION ": failed to make connection "
-                     "to backend: %s:%u", backend->hostname, backend->port);
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, APLOGNO(00866)
+                      "failed to make connection to backend: %s:%u",
+                      backend->hostname, backend->port);
         status = HTTP_SERVICE_UNAVAILABLE;
         goto cleanup;
     }
@@ -557,8 +586,8 @@ static void *create_scgi_config(apr_pool_t *p, char *dummy)
 {
     scgi_config *conf=apr_palloc(p, sizeof(*conf));
 
-    conf->sendfile = NULL;
-    conf->internal_redirect = -1;
+    conf->sendfile = NULL; /* === default (off) */
+    conf->internal_redirect = NULL; /* === default (on) */
 
     return conf;
 }
@@ -569,7 +598,7 @@ static void *merge_scgi_config(apr_pool_t *p, void *base_, void *add_)
     scgi_config *base=base_, *add=add_, *conf=apr_palloc(p, sizeof(*conf));
 
     conf->sendfile = add->sendfile ? add->sendfile: base->sendfile;
-    conf->internal_redirect =   (add->internal_redirect != -1)
+    conf->internal_redirect = add->internal_redirect
                               ? add->internal_redirect
                               : base->internal_redirect;
     return conf;
@@ -594,16 +623,33 @@ static const char *scgi_set_send_file(cmd_parms *cmd, void *mconfig,
 }
 
 
+static const char *scgi_set_internal_redirect(cmd_parms *cmd, void *mconfig,
+                                              const char *arg)
+{
+    scgi_config *conf = mconfig;
+
+    if (!strcasecmp(arg, "Off")) {
+        conf->internal_redirect = scgi_internal_redirect_off;
+    }
+    else if (!strcasecmp(arg, "On")) {
+        conf->internal_redirect = scgi_internal_redirect_on;
+    }
+    else {
+        conf->internal_redirect = arg;
+    }
+    return NULL;
+}
+
+
 static const command_rec scgi_cmds[] =
 {
     AP_INIT_TAKE1("ProxySCGISendfile", scgi_set_send_file, NULL,
                   RSRC_CONF|ACCESS_CONF,
-                  "The name of the X-Sendfile peudo response header or "
+                  "The name of the X-Sendfile pseudo response header or "
                   "On or Off"),
-    AP_INIT_FLAG("ProxySCGIInternalRedirect", ap_set_flag_slot,
-                 (void*)APR_OFFSETOF(scgi_config, internal_redirect),
-                 RSRC_CONF|ACCESS_CONF,
-                 "Off if internal redirect responses should not be accepted"),
+    AP_INIT_TAKE1("ProxySCGIInternalRedirect", scgi_set_internal_redirect, NULL,
+                  RSRC_CONF|ACCESS_CONF,
+                  "The name of the pseudo response header or On or Off"),
     {NULL}
 };
 
@@ -617,7 +663,7 @@ static void register_hooks(apr_pool_t *p)
 }
 
 
-module AP_MODULE_DECLARE_DATA proxy_scgi_module = {
+AP_DECLARE_MODULE(proxy_scgi) = {
     STANDARD20_MODULE_STUFF,
     create_scgi_config,  /* create per-directory config structure */
     merge_scgi_config,   /* merge per-directory config structures */
