@@ -1,9 +1,9 @@
-/* Copyright 1999-2005 The Apache Software Foundation or its licensors, as
- * applicable.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -15,7 +15,6 @@
  */
 
 #include "ap_config.h"
-#define CORE_PRIVATE
 #include "httpd.h"
 #include "http_config.h"
 #include "http_main.h"
@@ -51,182 +50,13 @@
 #include <sys/prctl.h>
 #endif
 
-unixd_config_rec unixd_config;
+unixd_config_rec ap_unixd_config;
 
-/* Set group privileges.
- *
- * Note that we use the username as set in the config files, rather than
- * the lookup of to uid --- the same uid may have multiple passwd entries,
- * with different sets of groups for each.
- */
+APLOG_USE_MODULE(core);
 
-static int set_group_privs(void)
-{
-    if (!geteuid()) {
-	const char *name;
-
-	/* Get username if passed as a uid */
-
-	if (unixd_config.user_name[0] == '#') {
-	    struct passwd *ent;
-	    uid_t uid = atoi(&unixd_config.user_name[1]);
-
-	    if ((ent = getpwuid(uid)) == NULL) {
-		ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-			 "getpwuid: couldn't determine user name from uid %u, "
-			 "you probably need to modify the User directive",
-			 (unsigned)uid);
-		return -1;
-	    }
-
-	    name = ent->pw_name;
-	}
-	else
-	    name = unixd_config.user_name;
-
-#if !defined(OS2) && !defined(TPF)
-	/* OS/2 and TPF don't support groups. */
-
-	/*
-	 * Set the GID before initgroups(), since on some platforms
-	 * setgid() is known to zap the group list.
-	 */
-	if (setgid(unixd_config.group_id) == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-			"setgid: unable to set group id to Group %u",
-			(unsigned)unixd_config.group_id);
-	    return -1;
-	}
-
-	/* Reset `groups' attributes. */
-
-	if (initgroups(name, unixd_config.group_id) == -1) {
-	    ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-			"initgroups: unable to set groups for User %s "
-			"and Group %u", name, (unsigned)unixd_config.group_id);
-	    return -1;
-	}
-#endif /* !defined(OS2) && !defined(TPF) */
-    }
-    return 0;
-}
-
-
-AP_DECLARE(int) unixd_setup_child(void)
-{
-    if (set_group_privs()) {
-	return -1;
-    }
-#ifdef MPE
-    /* Only try to switch if we're running as MANAGER.SYS */
-    if (geteuid() == 1 && unixd_config.user_id > 1) {
-	GETPRIVMODE();
-	if (setuid(unixd_config.user_id) == -1) {
-	    GETUSERMODE();
-	    ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-			"setuid: unable to change to uid: %ld",
-                        (long) unixd_config.user_id);
-	    exit(1);
-	}
-	GETUSERMODE();
-    }
-#else
-    /* Only try to switch if we're running as root */
-    if (!geteuid() && (
-#ifdef _OSD_POSIX
-	os_init_job_environment(server_conf, unixd_config.user_name, one_process) != 0 || 
-#endif
-	setuid(unixd_config.user_id) == -1)) {
-	ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-		    "setuid: unable to change to uid: %ld",
-                    (long) unixd_config.user_id);
-	return -1;
-    }
-#if defined(HAVE_PRCTL) && defined(PR_SET_DUMPABLE) 
-    /* this applies to Linux 2.4+ */
-#ifdef AP_MPM_WANT_SET_COREDUMPDIR
-    if (ap_coredumpdir_configured) {
-        if (prctl(PR_SET_DUMPABLE, 1)) {
-            ap_log_error(APLOG_MARK, APLOG_ALERT, errno, NULL,
-                         "set dumpable failed - this child will not coredump"
-                         " after software errors");
-        }
-    }
-#endif
-#endif
-#endif
-    return 0;
-}
-
-
-AP_DECLARE(const char *) unixd_set_user(cmd_parms *cmd, void *dummy, 
-                                        const char *arg)
-{
-    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
-    if (err != NULL) {
-        return err;
-    }
-
-    unixd_config.user_name = arg;
-    unixd_config.user_id = ap_uname2id(arg);
-#if !defined (BIG_SECURITY_HOLE) && !defined (OS2)
-    if (unixd_config.user_id == 0) {
-	return "Error:\tApache has not been designed to serve pages while\n"
-		"\trunning as root.  There are known race conditions that\n"
-		"\twill allow any local user to read any file on the system.\n"
-		"\tIf you still desire to serve pages as root then\n"
-		"\tadd -DBIG_SECURITY_HOLE to the CFLAGS env variable\n"
-		"\tand then rebuild the server.\n"
-		"\tIt is strongly suggested that you instead modify the User\n"
-		"\tdirective in your httpd.conf file to list a non-root\n"
-		"\tuser.\n";
-    }
-#endif
-
-    return NULL;
-}
-
-AP_DECLARE(const char *) unixd_set_group(cmd_parms *cmd, void *dummy, 
-                                         const char *arg)
-{
-    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
-    if (err != NULL) {
-        return err;
-    }
-
-    unixd_config.group_id = ap_gname2id(arg);
-
-    return NULL;
-}
-
-AP_DECLARE(void) unixd_pre_config(apr_pool_t *ptemp)
-{
-    apr_finfo_t wrapper;
-
-    unixd_config.user_name = DEFAULT_USER;
-    unixd_config.user_id = ap_uname2id(DEFAULT_USER);
-    unixd_config.group_id = ap_gname2id(DEFAULT_GROUP);
-
-    /* Check for suexec */
-    unixd_config.suexec_enabled = 0;
-    if ((apr_stat(&wrapper, SUEXEC_BIN, 
-                  APR_FINFO_NORM, ptemp)) != APR_SUCCESS) {
-        return;
-    }
-
-/* since APR 0.9.5 */
-#ifdef APR_USETID
-    if ((wrapper.protection & APR_USETID) && wrapper.user == 0) {
-#endif
-        unixd_config.suexec_enabled = 1;
-#ifdef APR_USETID
-    }
-#endif
-}
-
-
-AP_DECLARE(void) unixd_set_rlimit(cmd_parms *cmd, struct rlimit **plimit, 
-                           const char *arg, const char * arg2, int type)
+AP_DECLARE(void) ap_unixd_set_rlimit(cmd_parms *cmd, struct rlimit **plimit,
+                                     const char *arg,
+                                     const char * arg2, int type)
 {
 #if (defined(RLIMIT_CPU) || defined(RLIMIT_DATA) || defined(RLIMIT_VMEM) || defined(RLIMIT_NPROC) || defined(RLIMIT_AS)) && APR_HAVE_STRUCT_RLIMIT && APR_HAVE_GETRLIMIT
     char *str;
@@ -239,12 +69,12 @@ AP_DECLARE(void) unixd_set_rlimit(cmd_parms *cmd, struct rlimit **plimit,
     limit = *plimit;
     if ((getrlimit(type, limit)) != 0)  {
         *plimit = NULL;
-        ap_log_error(APLOG_MARK, APLOG_ERR, errno, cmd->server,
+        ap_log_error(APLOG_MARK, APLOG_ERR, errno, cmd->server, APLOGNO(02172)
                      "%s: getrlimit failed", cmd->cmd->name);
         return;
     }
 
-    if ((str = ap_getword_conf(cmd->pool, &arg))) {
+    if (*(str = ap_getword_conf(cmd->temp_pool, &arg)) != '\0') {
         if (!strcasecmp(str, "max")) {
             cur = limit->rlim_max;
         }
@@ -253,22 +83,25 @@ AP_DECLARE(void) unixd_set_rlimit(cmd_parms *cmd, struct rlimit **plimit,
         }
     }
     else {
-        ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server, APLOGNO(02173)
                      "Invalid parameters for %s", cmd->cmd->name);
         return;
     }
 
-    if (arg2 && (str = ap_getword_conf(cmd->pool, &arg2))) {
+    if (arg2 && (*(str = ap_getword_conf(cmd->temp_pool, &arg2)) != '\0')) {
         max = atol(str);
     }
 
     /* if we aren't running as root, cannot increase max */
     if (geteuid()) {
         limit->rlim_cur = cur;
-        if (max) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
+        if (max && (max > limit->rlim_max)) {
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server, APLOGNO(02174)
                          "Must be uid 0 to raise maximum %s", cmd->cmd->name);
         }
+        else if (max) {
+            limit->rlim_max = max;
+        }
     }
     else {
         if (cur) {
@@ -280,7 +113,7 @@ AP_DECLARE(void) unixd_set_rlimit(cmd_parms *cmd, struct rlimit **plimit,
     }
 #else
 
-    ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
+    ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server, APLOGNO(02175)
                  "Platform does not support rlimit for %s", cmd->cmd->name);
 #endif
 }
@@ -305,7 +138,7 @@ static apr_status_t ap_unix_create_privileged_process(
     char *execuser, *execgroup;
     const char *argv0;
 
-    if (!unixd_config.suexec_enabled) {
+    if (!ap_unixd_config.suexec_enabled) {
         return apr_proc_create(newproc, progname, args, env, attr, p);
     }
 
@@ -332,11 +165,8 @@ static apr_status_t ap_unix_create_privileged_process(
     }
 
     i = 0;
-    if (args) {
-        while (args[i]) {
-            i++;
-	    }
-    }
+    while (args[i])
+        i++;
     /* allocate space for 4 new args, the input args, and a null terminator */
     newargs = apr_palloc(p, sizeof(char *) * (i + 4));
     newprogname = SUEXEC_BIN;
@@ -393,7 +223,7 @@ static apr_lockmech_e proc_mutex_mech(apr_proc_mutex_t *pmutex)
     return APR_LOCK_DEFAULT;
 }
 
-AP_DECLARE(apr_status_t) unixd_set_proc_mutex_perms(apr_proc_mutex_t *pmutex)
+AP_DECLARE(apr_status_t) ap_unixd_set_proc_mutex_perms(apr_proc_mutex_t *pmutex)
 {
     if (!geteuid()) {
         apr_lockmech_e mech = proc_mutex_mech(pmutex);
@@ -411,11 +241,11 @@ AP_DECLARE(apr_status_t) unixd_set_proc_mutex_perms(apr_proc_mutex_t *pmutex)
             };
 #endif
             union semun ick;
-            struct semid_ds buf;
-        
+            struct semid_ds buf = { { 0 } };
+
             apr_os_proc_mutex_get(&ospmutex, pmutex);
-            buf.sem_perm.uid = unixd_config.user_id;
-            buf.sem_perm.gid = unixd_config.group_id;
+            buf.sem_perm.uid = ap_unixd_config.user_id;
+            buf.sem_perm.gid = ap_unixd_config.group_id;
             buf.sem_perm.mode = 0600;
             ick.buf = &buf;
             if (semctl(ospmutex.crossproc, 0, IPC_SET, ick) < 0) {
@@ -430,7 +260,7 @@ AP_DECLARE(apr_status_t) unixd_set_proc_mutex_perms(apr_proc_mutex_t *pmutex)
             const char *lockfile = apr_proc_mutex_lockfile(pmutex);
 
             if (lockfile) {
-                if (chown(lockfile, unixd_config.user_id,
+                if (chown(lockfile, ap_unixd_config.user_id,
                           -1 /* no gid change */) < 0) {
                     return errno;
                 }
@@ -446,28 +276,43 @@ AP_DECLARE(apr_status_t) unixd_set_proc_mutex_perms(apr_proc_mutex_t *pmutex)
     return APR_SUCCESS;
 }
 
-AP_DECLARE(apr_status_t) unixd_set_global_mutex_perms(apr_global_mutex_t *gmutex)
+AP_DECLARE(apr_status_t) ap_unixd_set_global_mutex_perms(apr_global_mutex_t *gmutex)
 {
 #if !APR_PROC_MUTEX_IS_GLOBAL
     apr_os_global_mutex_t osgmutex;
     apr_os_global_mutex_get(&osgmutex, gmutex);
-    return unixd_set_proc_mutex_perms(osgmutex.proc_mutex);
+    return ap_unixd_set_proc_mutex_perms(osgmutex.proc_mutex);
 #else  /* APR_PROC_MUTEX_IS_GLOBAL */
     /* In this case, apr_proc_mutex_t and apr_global_mutex_t are the same. */
-    return unixd_set_proc_mutex_perms(gmutex);
+    return ap_unixd_set_proc_mutex_perms(gmutex);
 #endif /* APR_PROC_MUTEX_IS_GLOBAL */
 }
 
-AP_DECLARE(apr_status_t) unixd_accept(void **accepted, ap_listen_rec *lr,
-                                        apr_pool_t *ptrans)
+AP_DECLARE(apr_status_t) ap_unixd_accept(void **accepted, ap_listen_rec *lr,
+                                         apr_pool_t *ptrans)
 {
     apr_socket_t *csd;
     apr_status_t status;
+#ifdef _OSD_POSIX
+    int sockdes;
+#endif
 
     *accepted = NULL;
-    status = apr_accept(&csd, lr->sd, ptrans);
-    if (status == APR_SUCCESS) { 
+    status = apr_socket_accept(&csd, lr->sd, ptrans);
+    if (status == APR_SUCCESS) {
         *accepted = csd;
+#ifdef _OSD_POSIX
+        apr_os_sock_get(&sockdes, csd);
+        if (sockdes >= FD_SETSIZE) {
+            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, ap_server_conf, APLOGNO(02176)
+                         "new file descriptor %d is too large; you probably need "
+                         "to rebuild Apache with a larger FD_SETSIZE "
+                         "(currently %d)",
+                         sockdes, FD_SETSIZE);
+            apr_socket_close(csd);
+            return APR_EINTR;
+        }
+#endif
         return APR_SUCCESS;
     }
 
@@ -566,28 +411,280 @@ AP_DECLARE(apr_status_t) unixd_accept(void **accepted, ap_listen_rec *lr,
              * Ben Hyde noted that temporary ENETDOWN situations
              * occur in mobile IP.
              */
-            ap_log_error(APLOG_MARK, APLOG_EMERG, status, ap_server_conf,
-                         "apr_accept: giving up.");
+            ap_log_error(APLOG_MARK, APLOG_EMERG, status, ap_server_conf, APLOGNO(02177)
+                         "apr_socket_accept: giving up.");
             return APR_EGENERAL;
 #endif /*ENETDOWN*/
 
-#ifdef TPF
-        case EINACT:
-            ap_log_error(APLOG_MARK, APLOG_EMERG, status, ap_server_conf,
-                         "offload device inactive");
-            return APR_EGENERAL;
-            break;
         default:
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, ap_server_conf,
-                         "select/accept error (%d)", status);
+            /* If the socket has been closed in ap_close_listeners()
+             * by the restart/stop action, we may get EBADF.
+             * Do not print an error in this case.
+             */
+            if (!lr->active) {
+                ap_log_error(APLOG_MARK, APLOG_DEBUG, status, ap_server_conf, APLOGNO(02178)
+                             "apr_socket_accept failed for inactive listener");
+                return status;
+            }
+            ap_log_error(APLOG_MARK, APLOG_ERR, status, ap_server_conf, APLOGNO(02179)
+                         "apr_socket_accept: (client socket)");
             return APR_EGENERAL;
-#else
-        default:
-            ap_log_error(APLOG_MARK, APLOG_ERR, status, ap_server_conf,
-                         "apr_accept: (client socket)");
-            return APR_EGENERAL;
-#endif
     }
     return status;
 }
 
+
+/* Unixes MPMs' */
+
+static ap_unixd_mpm_retained_data *retained_data = NULL;
+AP_DECLARE(ap_unixd_mpm_retained_data *) ap_unixd_mpm_get_retained_data()
+{
+    if (!retained_data) {
+        retained_data = ap_retained_data_create("ap_unixd_mpm_retained_data",
+                                                sizeof(*retained_data));
+        retained_data->mpm_state = AP_MPMQ_STARTING;
+    }
+    return retained_data;
+}
+
+static void sig_term(int sig)
+{
+    retained_data->mpm_state = AP_MPMQ_STOPPING;
+    if (retained_data->shutdown_pending
+            && (retained_data->is_ungraceful
+                || sig == AP_SIG_GRACEFUL_STOP)) {
+        /* Already handled */
+        return;
+    }
+
+    retained_data->shutdown_pending = 1;
+    if (sig != AP_SIG_GRACEFUL_STOP) {
+        retained_data->is_ungraceful = 1;
+    }
+}
+
+static void sig_restart(int sig)
+{
+    retained_data->mpm_state = AP_MPMQ_STOPPING;
+    if (retained_data->restart_pending
+            && (retained_data->is_ungraceful
+                || sig == AP_SIG_GRACEFUL)) {
+        /* Already handled */
+        return;
+    }
+
+    retained_data->restart_pending = 1;
+    if (sig != AP_SIG_GRACEFUL) {
+        retained_data->is_ungraceful = 1;
+    }
+}
+
+static apr_status_t unset_signals(void *unused)
+{
+    retained_data->shutdown_pending = retained_data->restart_pending = 0;
+    retained_data->was_graceful = !retained_data->is_ungraceful;
+    retained_data->is_ungraceful = 0;
+
+    return APR_SUCCESS;
+}
+
+AP_DECLARE(void) ap_unixd_mpm_set_signals(apr_pool_t *pconf, int one_process)
+{
+#ifndef NO_USE_SIGACTION
+    struct sigaction sa;
+#endif
+
+    /* Signals' handlers depend on retained data */
+    (void)ap_unixd_mpm_get_retained_data();
+
+#ifndef NO_USE_SIGACTION
+    memset(&sa, 0, sizeof sa);
+    sigemptyset(&sa.sa_mask);
+
+#ifdef SIGPIPE
+    sa.sa_handler = SIG_IGN;
+    if (sigaction(SIGPIPE, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00269)
+                     "sigaction(SIGPIPE)");
+#endif
+#ifdef SIGXCPU
+    sa.sa_handler = SIG_DFL;
+    if (sigaction(SIGXCPU, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00267)
+                     "sigaction(SIGXCPU)");
+#endif
+#ifdef SIGXFSZ
+    /* For systems following the LFS standard, ignoring SIGXFSZ allows
+     * a write() beyond the 2GB limit to fail gracefully with E2BIG
+     * rather than terminate the process. */
+    sa.sa_handler = SIG_IGN;
+    if (sigaction(SIGXFSZ, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00268)
+                     "sigaction(SIGXFSZ)");
+#endif
+
+    sa.sa_handler = sig_term;
+    if (sigaction(SIGTERM, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00264)
+                     "sigaction(SIGTERM)");
+#ifdef SIGINT
+    if (sigaction(SIGINT, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00266)
+                     "sigaction(SIGINT)");
+#endif
+#ifdef AP_SIG_GRACEFUL_STOP
+    if (sigaction(AP_SIG_GRACEFUL_STOP, &sa, NULL) < 0)
+        ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00265)
+                     "sigaction(" AP_SIG_GRACEFUL_STOP_STRING ")");
+#endif
+
+    /* Don't catch restart signals in ONE_PROCESS mode :) */
+    if (!one_process) {
+        sa.sa_handler = sig_restart;
+        if (sigaction(SIGHUP, &sa, NULL) < 0)
+            ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00270)
+                         "sigaction(SIGHUP)");
+        if (sigaction(AP_SIG_GRACEFUL, &sa, NULL) < 0)
+            ap_log_error(APLOG_MARK, APLOG_WARNING, errno, ap_server_conf, APLOGNO(00271)
+                         "sigaction(" AP_SIG_GRACEFUL_STRING ")");
+    }
+
+#else  /* NO_USE_SIGACTION */
+
+#ifdef SIGPIPE
+    apr_signal(SIGPIPE, SIG_IGN);
+#endif /* SIGPIPE */
+#ifdef SIGXCPU
+    apr_signal(SIGXCPU, SIG_DFL);
+#endif /* SIGXCPU */
+#ifdef SIGXFSZ
+    apr_signal(SIGXFSZ, SIG_IGN);
+#endif /* SIGXFSZ */
+
+    apr_signal(SIGTERM, sig_term);
+#ifdef AP_SIG_GRACEFUL_STOP
+    apr_signal(AP_SIG_GRACEFUL_STOP, sig_term);
+#endif /* AP_SIG_GRACEFUL_STOP */
+
+    if (!one_process) {
+        /* Don't restart in ONE_PROCESS mode :) */
+#ifdef SIGHUP
+        apr_signal(SIGHUP, sig_restart);
+#endif /* SIGHUP */
+#ifdef AP_SIG_GRACEFUL
+        apr_signal(AP_SIG_GRACEFUL, sig_restart);
+#endif /* AP_SIG_GRACEFUL */
+    }
+
+#endif /* NO_USE_SIGACTION */
+
+    apr_pool_cleanup_register(pconf, NULL, unset_signals,
+                              apr_pool_cleanup_null);
+}
+
+
+#ifdef _OSD_POSIX
+
+#include "apr_lib.h"
+
+#define USER_LEN 8
+
+typedef enum
+{
+    bs2_unknown,     /* not initialized yet. */
+    bs2_noFORK,      /* no fork() because -X flag was specified */
+    bs2_FORK,        /* only fork() because uid != 0 */
+    bs2_UFORK        /* Normally, ufork() is used to switch identities. */
+} bs2_ForkType;
+
+static bs2_ForkType forktype = bs2_unknown;
+
+/* Determine the method for forking off a child in such a way as to
+ * set both the POSIX and BS2000 user id's to the unprivileged user.
+ */
+static bs2_ForkType os_forktype(int one_process)
+{
+    /* have we checked the OS version before? If yes return the previous
+     * result - the OS release isn't going to change suddenly!
+     */
+    if (forktype == bs2_unknown) {
+        /* not initialized yet */
+
+        /* No fork if the one_process option was set */
+        if (one_process) {
+            forktype = bs2_noFORK;
+        }
+        /* If the user is unprivileged, use the normal fork() only. */
+        else if (getuid() != 0) {
+            forktype = bs2_FORK;
+        }
+        else
+            forktype = bs2_UFORK;
+    }
+    return forktype;
+}
+
+
+
+/* This routine complements the setuid() call: it causes the BS2000 job
+ * environment to be switched to the target user's user id.
+ * That is important if CGI scripts try to execute native BS2000 commands.
+ */
+int os_init_job_environment(server_rec *server, const char *user_name, int one_process)
+{
+    bs2_ForkType            type = os_forktype(one_process);
+
+    /* We can be sure that no change to uid==0 is possible because of
+     * the checks in http_core.c:set_user()
+     */
+
+    if (one_process) {
+
+        type = forktype = bs2_noFORK;
+
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, server, APLOGNO(02180)
+                     "The debug mode of Apache should only "
+                     "be started by an unprivileged user!");
+        return 0;
+    }
+
+    return 0;
+}
+
+/* BS2000 requires a "special" version of fork() before a setuid() call */
+pid_t os_fork(const char *user)
+{
+    pid_t pid;
+    char  username[USER_LEN+1];
+
+    switch (os_forktype(0)) {
+
+      case bs2_FORK:
+        pid = fork();
+        break;
+
+      case bs2_UFORK:
+        apr_cpystrn(username, user, sizeof username);
+
+        /* Make user name all upper case - for some versions of ufork() */
+        ap_str_toupper(username);
+
+        pid = ufork(username);
+        if (pid == -1 && errno == EPERM) {
+            ap_log_error(APLOG_MARK, APLOG_EMERG, errno, ap_server_conf,
+                         APLOGNO(02181) "ufork: Possible mis-configuration "
+                         "for user %s - Aborting.", user);
+            exit(1);
+        }
+        break;
+
+      default:
+        pid = 0;
+        break;
+    }
+
+    return pid;
+}
+
+#endif /* _OSD_POSIX */
+
